import os
import cocotb
import functools

from cocotb.triggers import (
    Timer,
    Edge,
)

from cocotb_test import simulator as cocotb_simulator

from cohdl import Entity, Port, BitVector, Signed, Unsigned
from cohdl import std

from ._proxy_port import ProxyPort
from ._base_simulation import _GenericParams, _BaseSimulator


class Task:
    def __init__(self, handle):
        self._handle = handle

    async def join(self):
        await self._handle.join()


class Simulator(_BaseSimulator):

    def _init_impl(self, p: _GenericParams, cocotb_extra_args=None):
        from cohdl_sim._build_cache import write_cache_file, load_cache_file

        # This code is evaluated twice. Once in normal user code
        # to setup the test environment and again from another process
        # started by cocotb_simulator.run().
        # Use an environment variable to determine current mode.
        if os.getenv("COHDLSIM_TEST_RUNNING") is None:
            # running in normal user code
            # run CoHDL design into VHDL code and
            # start cocotb simulator

            top_name = p.entity._cohdl_info.name

            if not p.use_build_cache:
                lib = std.VhdlCompiler.to_vhdl_library(p.entity)
                vhdl_sources = (
                    p.extra_vhdl_files
                    + lib.write_dir(p.vhdl_dir)
                    + p.extra_vhdl_files_post
                )

                write_cache_file(p.cache_file, p.entity, vhdl_sources=vhdl_sources)
            else:
                cache_content = load_cache_file(p.cache_file, p.entity)
                vhdl_sources = cache_content.vhdl_sources

            # cocotb_simulator.run() requires the module name
            # of the Python file containing the test benches

            import inspect
            import pathlib

            filename = inspect.stack()[2].filename
            filename = pathlib.Path(filename).stem

            cocotb_extra_args = {} if cocotb_extra_args is None else cocotb_extra_args

            cocotb_simulator.run(
                simulator=p.simulator,
                sim_args=p.sim_args,
                sim_build=p.sim_dir,
                vhdl_sources=vhdl_sources,
                toplevel=top_name.lower(),
                module=filename,
                extra_env={"COHDLSIM_TEST_RUNNING": "True", **p.extra_env},
                **cocotb_extra_args,
            )
        else:
            # running in simulator process
            # initialize members used by Simulator.test

            # instantiate entity to generate dynamic ports
            p.entity(_cohdl_instantiate_only=True)
            self._dut = None

    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        simulator: str = "ghdl",
        sim_args: list[str] | None = None,
        sim_dir: str = "sim",
        vhdl_dir: str = "vhdl",
        cast_vectors=None,
        extra_env: dict[str, str] | None = None,
        extra_vhdl_files: list[str] = None,
        use_build_cache: bool = False,
        cocotb_extra_args: dict[str, str] | None = None,
    ):
        p = _GenericParams(
            entity=entity,
            build_dir=build_dir,
            simulator=simulator,
            sim_args=sim_args,
            sim_dir=sim_dir,
            vhdl_dir=vhdl_dir,
            cast_vectors=cast_vectors,
            extra_env=extra_env,
            extra_vhdl_files=extra_vhdl_files,
            use_build_cache=use_build_cache,
        )

        super().__init__(p)
        self._init_impl(p, cocotb_extra_args=cocotb_extra_args)

    async def wait(self, duration: std.Duration, /):
        await Timer(int(duration.picoseconds()), units="ps")

    async def delta_step(self):
        await Timer(1, units="step")

    async def rising_edge(self, signal: ProxyPort, /):
        await signal._rising_edge()

    async def falling_edge(self, signal: ProxyPort, /):
        await signal._falling_edge()

    async def any_edge(self, signal: ProxyPort, /):
        await signal._edge()

    async def clock_cycles(self, signal: ProxyPort, num_cycles: int, rising=True):
        if isinstance(signal, std.Clock):
            signal = signal.signal()

        await signal._clock_cycles(num_cycles, rising)

    async def value_change(self, signal: ProxyPort, /):
        await Edge(signal._cocotb_port)

    async def value_true(self, signal: ProxyPort, /):
        while not signal:
            await signal._edge()

    async def value_false(self, signal: ProxyPort, /):
        while signal:
            await signal._edge()

    #
    #

    async def start(self, coro, /):
        return Task(await cocotb.start(coro))

    def start_soon(self, coro, /):
        return Task(cocotb.start_soon(coro))

    def gen_clock(
        self, clk, period_or_frequency: std.Duration = None, /, start_state=False
    ):
        if isinstance(clk, std.Clock):
            if period_or_frequency is None:
                period_or_frequency = clk.frequency()

            clk = clk.signal()

        assert isinstance(period_or_frequency, (std.Frequency, std.Duration))

        period = period_or_frequency.period()

        half = int(period.picoseconds()) // 2

        async def thread():
            nonlocal clk
            while True:
                clk <<= start_state
                await Timer(half, units="ps")
                clk <<= not start_state
                await Timer(half, units="ps")

        self.start_soon(thread())

    def get_dut(self):
        assert (
            self._dut is not None
        ), "get_dut may only be called from a testbench function running in cocotb"
        return self._dut

    def test(self, testbench, /):
        @cocotb.test()
        @functools.wraps(testbench)
        async def helper(dut):
            self._ports = {}
            self._input_ports = {}
            self._output_ports = {}
            self._inout_ports = {}

            self._dut = dut
            entity = self._params.entity
            entity_name = entity.__name__

            class EntityProxy(entity):
                def __new__(cls):
                    return object.__new__(cls)

                def __init__(self):
                    pass

                def __str__(self):
                    return entity_name

                def __repr__(self):
                    return entity_name

            for name, port in entity._cohdl_info.ports.items():

                port_bv = self._params.cast_vectors

                if port_bv is not None:
                    port_type = type(Port.decay(port))

                    if issubclass(port_type, BitVector) and not (
                        issubclass(port_type, (Signed, Unsigned))
                    ):
                        if port_bv is Unsigned:
                            port = port.unsigned
                        elif port_bv is Signed:
                            port = port.signed
                        else:
                            raise AssertionError(
                                f"invalid default vector port type {port_bv}"
                            )

                proxy = ProxyPort(port, None, getattr(dut, name))
                setattr(EntityProxy, name, proxy)

                self._ports[name] = proxy

                match port.direction():
                    case Port.Direction.INPUT:
                        self._input_ports[name] = proxy
                    case Port.Direction.OUTPUT:
                        self._output_ports[name] = proxy
                    case Port.Direction.INOUT:
                        self._inout_ports[name] = proxy

            # first delta step to load default values
            await Timer(1, units="step")
            # await self.delta_step()

            # use proxy entity instead of cocotb dut
            # in testbench function
            await testbench(EntityProxy())

        return helper

    def freeze(self, port: ProxyPort, /):
        port.freeze()

    def release(self, port: ProxyPort, /):
        port.release()



import os
import shutil
import subprocess
from pathlib import Path


def run_command(command, *args):
    cmd_string = f"{command} {' '.join(str(arg) for arg in args)}"

    result = subprocess.run([command, *args], stdout=subprocess.PIPE)

    if result.returncode != 0:
        print(result.stderr)
        print(result.stdout)
        raise AssertionError(f"command failed: {cmd_string}")

    return result.stdout.decode()


def prepare_ghdl_simulation(
    vhdl_sources: list[str], top_module: str, build_dir=Path, copy_files=False
) -> Path:
    status, _ = subprocess.getstatusoutput("ghdl-gcc --version")
    assert status == 0, "the ghdl_sim simulator requires the ghdl backend ghdl-gcc"

    vhdl_names = []

    for source_path in vhdl_sources:
        file_name = Path(source_path).name
        vhdl_names.append(file_name)

        if copy_files:
            target_path = build_dir / file_name
            shutil.copyfile(source_path, target_path)
        else:
            assert Path(source_path).parent == build_dir

    curdir = Path(os.curdir).absolute()

    try:
        os.chdir(build_dir)
        run_command("ghdl-gcc", "-a", *vhdl_names)
        run_command("ghdl-gcc", "--bind", top_module)
        list_link = run_command("ghdl-gcc", "--list-link", top_module).split()

        version_script_path = None
        filtered_list_link = []

        VERSION_SCRIPT_PREFIX = "-Wl,--version-script="

        for arg in list_link:
            if arg.startswith(VERSION_SCRIPT_PREFIX):
                version_script_path = Path(arg.removeprefix("-Wl,--version-script="))
            else:
                filtered_list_link.append(arg)

        # generate a modified version-script to make the symbol
        # `ghdl_main` globally visible
        local_version_script = "version-script.ver"

        with open(local_version_script, "w+") as local_file:
            with open(version_script_path) as original_file:
                for line in original_file:
                    print(line, file=local_file, end="")

                    if "global:" in line:
                        print("ghdl_main;", file=local_file)

        list_link = filtered_list_link + ["-Wl,--version-script=version-script.ver"]

        out_name = f"lib{top_module}.so"

        run_command("gcc", *list_link, "-Wl,-shared,-fPIC", f"-o{out_name}")

        return Path(out_name).absolute()
    finally:
        os.chdir(curdir)



from cohdl_sim_ghdl_interface import ObjHandle

from cohdl import Signal, Bit, BitVector, Unsigned, Signed
from cohdl_sim._base_proxy_port import _BaseProxyPort


class ProxyPort(_BaseProxyPort):
    def __init__(
        self, entity_port: Signal, root=None, ghdl_handle: ObjHandle = None, sim=None
    ):
        super().__init__(entity_port, root)

        # the remaining members are None unless this is the root port
        self._handle = ghdl_handle
        self._sim = sim

    def _load(self):
        val = self._handle.get_binstr()

        if issubclass(self._type, (Bit, BitVector)):
            self._Wrapped._assign(val.upper())
        else:
            raise AssertionError(f"type {type(self._type)} not supported")

    def _store(self):
        if isinstance(self._Wrapped, (Unsigned, Signed)):
            self._handle.put_binstr(str(self._Wrapped.bitvector))
        else:
            self._handle.put_binstr(str(self._Wrapped))

    def __await__(self):
        async def gen():
            return await self._root._sim.value_true(self)

        return gen().__await__()



from cohdl import BitVector, Signed, Unsigned, Signal, Port, TypeQualifierBase


decay = TypeQualifierBase.decay


def load_all(*args):
    for arg in args:
        if isinstance(arg, _BaseProxyPort):
            arg._load()


class _BaseProxyPort(TypeQualifierBase):
    def __init__(
        self,
        entity_port: Signal[BitVector],
        root=None,
    ):
        self._Wrapped = Port.decay(entity_port)
        self._type = type(self._Wrapped)
        self._root = self if root is None else root

    def decay(self):
        return Port.decay(self._Wrapped)

    def __call__(self):
        return self

    def _is_root(self):
        return self._root is self

    def _to_type(self, inp):
        return self._type(inp)

    def copy(self):
        self._load()
        return self._Wrapped.copy()

    @property
    def signed(self):
        assert issubclass(self._type, BitVector)

        if issubclass(self._type, Signed):
            return self

        result = type(self)(self._Wrapped.signed, self._root)
        result._tmp_signed_parent = self
        result._load = self._load
        result._store = self._store
        return result

    @signed.setter
    def signed(self, value):
        assert (
            value is self or value._tmp_signed_parent is self
        ), "direct assignment to .signed property not allowed use '<<=' operator"

    @property
    def unsigned(self):
        assert issubclass(self._type, BitVector)

        if issubclass(self._type, Unsigned):
            return self

        result = type(self)(self._Wrapped.unsigned, self._root)
        result._tmp_unsigned_parent = self
        result._load = self._load
        result._store = self._store
        return result

    @unsigned.setter
    def unsigned(self, value):
        assert (
            value is self or value._tmp_unsigned_parent is self
        ), "direct assignment to .unsigned property not allowed use '<<=' operator"

    @property
    def bitvector(self):
        assert issubclass(self._type, BitVector)

        if not issubclass(self._type, (Signed, Unsigned)):
            return self

        result = type(self)(self._Wrapped.bitvector, self._root)
        result._tmp_bitvector_parent
        result._load = self._load
        result._store = self._store
        return result

    @bitvector.setter
    def bitvector(self, value):
        assert (
            value is self or value._tmp_bitvector_parent is self
        ), "direct assignment to .bitvector property not allowed use '<<=' operator"

    def __getitem__(self, arg):
        assert issubclass(self._type, BitVector)

        if isinstance(arg, slice):
            assert isinstance(arg.start, int)
            assert isinstance(arg.stop, int)
            assert arg.step is None
            result = type(self)(self._Wrapped[arg], self._root)
        else:
            assert isinstance(arg, int)
            result = type(self)(self._Wrapped[arg], self._root)

        # Replace load and store methods with version
        # of root object. They always update all bits of
        # a port even if they are accessed via a slice.
        result._load = self._load
        result._store = self._store

        return result

    def __setitem__(self, arg, value):
        pass

    def __ilshift__(self, src):
        if isinstance(src, _BaseProxyPort):
            src._load()
            src = src._Wrapped

        self._Wrapped._assign(src)
        self._store()
        return self

    @property
    def next(self):
        raise AssertionError("reading from .next property not allowed")

    @next.setter
    def next(self, value):
        self <<= value

    @property
    def type(self):
        return self._type

    def __bool__(self):
        self._load()
        return self._Wrapped.__bool__()

    def __index__(self):
        self._load()
        return self._Wrapped.__index__()

    def __eq__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__eq__(other)

    def __gt__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__gt__(other)

    def __lt__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__lt__(other)

    def __ge__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__ge__(other)

    def __invert__(self):
        self._load()
        return self._Wrapped.__invert__()

    def __le__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__le__(other)

    def __add__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__add__(other)

    def __sub__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__sub__(other)

    def __and__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__and__(other)

    def __or__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__or__(other)

    def __xor__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__xor__(other)

    def __matmul__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__matmul__(other)

    def __radd__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__radd__(other)

    def __rsub__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__rsub__(other)

    def __rand__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__rand__(other)

    def __ror__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__ror__(other)

    def __rxor__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__rxor__(other)

    def __rmatmul__(self, other):
        load_all(self, other)
        other = decay(other)
        return self._Wrapped.__rmatmul__(other)

    def resize(self, *args, **kwargs):
        return decay(self).resize(*args, **kwargs)

    def __str__(self):
        self._load()
        return str(self._Wrapped)

    def __repr__(self):
        self._load()
        return repr(self._Wrapped)

    #
    # abstract methods
    #

    def _load(self):
        raise AssertionError("abstract method called")

    def _store(self):
        raise AssertionError("abstract method called")



from __future__ import annotations

from cohdl import std, BitVector, Unsigned
from cohdl_sim import Simulator

class Axi4Light:
    def __init__(self, sim: Simulator, con: std.axi.axi4_light.Axi4Light):
        """ """

    async def read(self, address: int | Unsigned) -> BitVector:
        """
        Send one read request, wait for the response and
        return the received data.
        """

    async def read_multiple(self, addresses: list[int | Unsigned]) -> list[BitVector]:
        """
        Send multiple read requests, wait for all responses and
        return them in a list.
        """

    async def write(self, address: int | Unsigned, value: BitVector):
        """ """

    async def write_multiple(self, addr_value: list[tuple[int | Unsigned, BitVector]]):
        """ """



from cohdl import Entity, Signal

from ._base_simulation import _BaseSimulator

class Simulator(_BaseSimulator):
    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        sim_dir: str = "sim",
        vhdl_dir: str = "vhdl",
        extra_vhdl_files: list[str] = None,
        extra_vhdl_files_post: list[str] = None,
        cast_vectors=None,
        simulator: str = "ghdl",
        sim_args: list[str] | None = None,
        extra_env: dict[str, str] | None = None,
        cocotb_extra_args: dict[str, str] | None = None,
    ):
        """
        Simulator takes a CoHDL entity, generates VHDL code from it
        and starts a cocotb test session.

        * `entity` tested entity type
        * `build_dir` simulation outputs are written to this directory
        * `sim_dir` name of directory inside build_dir that is used for simulator output
        * `vhdl_dir` name of directory inside build_dir where generated VHDL files are written
        * `extra_vhdl_files` list of paths to additional VHDL files
        * `extra_vhdl_files_post` like `extra_vhdl_files` but arguments are analyzed after CoHDL entities
        * `cast_vectors` when set to cohdl.Signed or cohdl.Unsigned all BitVector ports are converted
                            to the corresponding type

        The following arguments are forwarded to cocotb:

        * `simulator`
        * `sim_args`
        * `extra_env`
        * `cocotb_extra_args` (as additional keyword arguments)
        """

    def get_dut(self):
        """
        return the cocotb design under test object
        """

    def freeze(self, port: Signal, /):
        """
        freeze the signal in its current state
        """

    def release(self, port: Signal, /):
        """
        release a frozen signal
        """



"""
Alternative simulator backend, implemented using direct GHDL access.
Only available on Linux platforms.

By default cohdl_sim uses cocotb to simulate designs.

The direct GHDL simulator supports fewer features.
It cannot be combined with cocotb libraries.
The main advantage is, that the entire simulation is performed
in a single Python context. Debuggers can be used to step threw
testbench code without additional setup.
Unlike the cocotb backend the direct GHDL access is usable from
jupyter notebooks.
"""

import os
import subprocess

assert os.name == "posix", "the ghdl_sim simulator is only supported in Linux"

success, output = subprocess.getstatusoutput("ghdl-gcc --version")
assert success == 0, "ghdl_sim requires `ghdl` and the ghdl backend `ghdl-gcc`"

try:
    import cohdl_sim_ghdl_interface
except:
    raise AssertionError(
        "The internal module cohdl_sim_ghdl_interface is not installed.\n"
        "A possible reason for this error is, that the ghdl command was not available while `pip install cohdl_sim` was executed.\n"
        "Try to install ghdl and ghdl-gcc and then reinstall cohdl_sim.\n"
    )


from ._simulation import Simulator



import os

from abc import ABC, abstractmethod

from pathlib import Path

from cohdl import Entity
from cohdl import std
from cohdl import Null

from ._proxy_port import ProxyPort


class Task:
    async def join(self): ...


class _GenericParams:
    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        simulator: str = "ghdl",
        sim_args: list[str] = None,
        sim_dir: str = "sim",
        vhdl_dir: str = "sim",
        cast_vectors=None,
        extra_env: dict[str, str] | None = None,
        extra_vhdl_files: list[str] = None,
        extra_vhdl_files_post: list[str] = None,
        use_build_cache: bool = False,
    ):
        self.entity = entity
        self.build_dir = Path(build_dir)

        self.simulator = simulator
        self.sim_dir = self.build_dir / sim_dir
        self.vhdl_dir = self.build_dir / vhdl_dir

        self.sim_args = [] if sim_args is None else sim_args
        self.extra_env = {} if extra_env is None else extra_env
        self.extra_vhdl_files = [] if extra_vhdl_files is None else extra_vhdl_files
        self.extra_vhdl_files_post = (
            [] if extra_vhdl_files_post is None else extra_vhdl_files_post
        )

        self.cache_file = self.build_dir / ".build-cache.json"

        # use cache file requested and available, rebuild it otherwise
        self.use_build_cache = use_build_cache and self.cache_file.exists()

        self.cast_vectors = cast_vectors

        for dir in (self.build_dir, self.sim_dir, self.vhdl_dir):
            if not os.path.exists(dir):
                os.makedirs(dir, exist_ok=True)


class _BaseSimulator(ABC):
    def __init__(self, params: _GenericParams):
        self._params = params

    @abstractmethod
    def test(self, testbench, /): ...

    @abstractmethod
    async def wait(self, duration: std.Duration, /): ...

    @abstractmethod
    async def delta_step(self): ...

    @abstractmethod
    async def rising_edge(self, signal: ProxyPort, /): ...

    @abstractmethod
    async def falling_edge(self, signal: ProxyPort, /): ...

    @abstractmethod
    async def any_edge(self, signal: ProxyPort, /): ...

    @abstractmethod
    async def clock_cycles(
        self, signal: ProxyPort | std.Clock, num_cycles: int, rising=True
    ): ...

    @abstractmethod
    async def value_change(self, signal: ProxyPort, /): ...

    @abstractmethod
    async def value_true(self, signal: ProxyPort, /): ...

    @abstractmethod
    async def value_false(self, signal: ProxyPort, /): ...

    @abstractmethod
    async def start(self, coro, /) -> Task: ...

    @abstractmethod
    def start_soon(self, coro, /) -> Task: ...

    @abstractmethod
    def gen_clock(
        self, clk, period_or_frequency: std.Duration = None, /, start_state=False
    ): ...

    def init_inputs(self, init_val=Null, /):
        for port in self._input_ports.values():
            port <<= init_val

    def init_outputs(self, init_val=Null, /):
        for port in self._output_ports.values():
            port <<= init_val

    def init_inouts(self, init_val=Null, /):
        for port in self._inout_ports.values():
            port <<= init_val

    async def clock_edge(self, clk: std.Clock, /):
        Edge = std.Clock.Edge

        match clk.edge():
            case Edge.RISING:
                await self.rising_edge(clk.signal())
            case Edge.FALLING:
                await self.falling_edge(clk.signal())
            case Edge.BOTH:
                await self.any_edge(clk.signal())
            case _:
                raise AssertionError(f"invalid clock edge {clk.edge()}")

    async def reset_cond(self, reset: std.Reset, /):
        if reset.is_active_high():
            if reset.is_async():
                await self.value_true(reset.signal())
            else:
                await self.rising_edge(reset.signal())
        else:
            if reset.is_async():
                await self.value_false(reset.signal())
            else:
                await self.falling_edge(reset.signal())

    async def true_on_rising(self, clk: ProxyPort, cond, *, timeout: int | None = None):
        while True:
            await self.rising_edge(clk)

            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_on_falling(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.falling_edge(clk)

            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_on_clk(self, clk: std.Clock, cond, *, timeout: int | None = None):
        if clk.is_rising_edge():
            await self.true_on_rising(clk.signal(), cond, timeout=timeout)
        else:
            await self.true_on_falling(clk.signal(), cond, timeout=timeout)

    async def true_after_rising(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.rising_edge(clk)
            await self.delta_step()

            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_after_falling(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.falling_edge(clk)
            await self.delta_step()

            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_after_clk(self, clk: std.Clock, cond, *, timeout: int | None = None):
        if clk.is_rising_edge():
            await self.true_after_rising(clk.signal(), cond, timeout=timeout)
        else:
            await self.true_after_falling(clk.signal(), cond, timeout=timeout)



from cohdl import Entity, Bit, Unsigned, Port
from cohdl import std

from cohdl_sim.ghdl_sim import Simulator


class MyEntity(Entity):
    clk = Port.input(Bit)
    reset = Port.input(Bit)

    cnt = Port.output(Unsigned[8], default=0)

    def architecture(self):
        @std.sequential(std.Clock(self.clk), std.Reset(self.reset))
        def proc():
            self.cnt <<= self.cnt + 1


#
# test code for MyEntity
#

# Simulator forwards arguments to cocotb_test.simulator.run().
# This can be used to pass arguments to the simulator (default=ghdl).
# For example the flag --vcd generates a value dump.
sim = Simulator(MyEntity, sim_args=["--vcd=waveform.vcd"])


@sim.test
async def testbench_cnt(entity: MyEntity):
    entity.clk <<= False
    entity.reset <<= True

    # wait for 1 ns of simulation time
    await sim.wait(std.ns(1))
    entity.clk <<= True
    await sim.wait(std.ns(1))

    # reset complete after rising edge on entity.clk
    entity.reset <<= False
    entity.clk <<= False

    # generate 10 clock cycles and check that entity.cnt is incremented
    for cnt in range(0, 10):
        assert entity.cnt == cnt

        await sim.wait(std.ns(1))
        entity.clk <<= True
        await sim.wait(std.ns(1))
        entity.clk <<= False


@sim.test
async def testbench_gen_clk(entity: MyEntity):
    # gen_clk starts a parallel task that
    # generates a clock signal of the requested frequency
    sim.gen_clock(entity.clk, std.GHz(1))

    # reset design by setting the reset signal to true
    # and waiting for a rising edge on entity.clk
    entity.reset <<= True
    await sim.rising_edge(entity.clk)
    entity.reset <<= False

    # check entity.cnt for 10 clock cycles
    for cnt in range(0, 10):
        assert entity.cnt == cnt

        await sim.rising_edge(entity.clk)



from cohdl import Signal, Bit
from cohdl import std
from cohdl import Null

from ._proxy_port import ProxyPort

class Task:
    async def join(self):
        """
        Wait until the task is done
        """

class _BaseSimulator:
    def test(self, testbench, /):
        """
        decorator that turns coroutines into test benches
        """
        return testbench

    async def wait(self, duration: std.Duration, /) -> None:
        """
        wait for a given simulation duration
        """

    async def delta_step(self) -> None:
        """
        run simulation for a short time to update output ports
        """

    async def rising_edge(self, signal: Signal[Bit], /) -> None:
        """
        wait for rising edge on signal
        """

    async def falling_edge(self, signal: Signal[Bit], /) -> None:
        """
        wait for falling edge on signal
        """

    async def any_edge(self, signal: Signal[Bit], /) -> None:
        """
        wait for changes of signal
        """

    async def clock_cycles(
        self, signal: Signal[Bit] | std.Clock, num_cycles: int, rising=True
    ):
        """
        wait for a number of clock cycles,
        `rising` determines whether rising or falling edges are counted
        """

    async def value_change(self, signal: Signal, /) -> None:
        """
        wait until signal changes
        """

    async def value_true(self, signal: Signal, /) -> None:
        """
        wait until signal becomes truthy
        """

    async def value_false(self, signal: Signal, /) -> None:
        """
        wait until signal becomes falsy
        """

    async def start(self, coro, /) -> Task:
        """
        Run coro in parallel task.
        `coro` will start immediately (current task is suspended).
        """

    def start_soon(self, coro, /) -> Task:
        """
        Run coro in parallel task.
        `coro` will start once the current task is suspended.
        """

    def gen_clock(
        self,
        clk: Signal[Bit] | std.Clock,
        period_or_frequency: std.Duration | std.Frequency | None = None,
        /,
        start_state=False,
    ) -> None:
        """
        Start a parallel task that produces a clock signal
        with the specified period or frequency on `clk`.

        The `period_of_frequency` parameter is mandatory unless
        `clk` is a `std.Clock` and defines its own frequency.
        """

    def init_inputs(self, init_val=Null, /):
        """
        assign `init_val` to all input ports of the tested entity
        """

    def init_outputs(self, init_val=Null, /):
        """
        assign `init_val` to all output ports of the tested entity
        """

    def init_inouts(self, init_val=Null, /):
        """
        assign `init_val` to all inout ports of the tested entity
        """

    async def clock_edge(self, clk: std.Clock, /):
        """
        wait for active edge on the clock
        """

    async def reset_cond(self, reset: std.Reset, /) -> None:
        """
        wait until the reset condition becomes true
        """

    async def true_on_rising(
        self, clk: Signal[Bit], cond, *, timeout: int | None = None
    ) -> None:
        """
        Wait until cond is true after a rising edge of the clock signal.
        `cond` can be a port or a callable taking no arguments returning a boolean value.
        Raises an exception if the condition remains false for more than timeout rising edges.
        """

    async def true_on_falling(
        self, clk: Signal[Bit], cond, *, timeout: int | None = None
    ) -> None:
        """
        Wait until cond is true after a falling edge of the clock signal.
        `cond` can be a port or a callable taking no arguments returning a boolean value.
        Raises an exception if the condition remains false for more than timeout falling edges.
        """

    async def true_on_clk(self, clk: std.Clock, cond, *, timeout: int | None = None):
        """
        Wait until cond is true after a clock tick.
        `cond` can be a port or a callable taking no arguments returning a boolean value.
        Raises an exception if the condition remains false for more than timeout ticks.
        """

    async def true_after_rising(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.rising_edge(clk)
            await self.delta_step()

            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_after_falling(
        self, clk: ProxyPort, cond, *, timeout: int | None = None
    ):
        while True:
            await self.falling_edge(clk)
            await self.delta_step()

            if cond():
                return

            if timeout is not None:
                assert timeout != 0, "timeout while waiting for condition"
                timeout -= 1

    async def true_after_clk(self, clk: std.Clock, cond, *, timeout: int | None = None):
        if clk.is_rising_edge():
            await self.true_after_rising(clk.signal(), cond, timeout=timeout)
        else:
            await self.true_after_falling(clk.signal(), cond, timeout=timeout)



from __future__ import annotations

from cohdl import std, Null, Full, BitVector, Unsigned
from cohdl_sim import Simulator


class Axi4Light:
    def __init__(self, sim: Simulator, con: std.axi.axi4_light.Axi4Light):
        self._con = con
        self._sim = sim

    async def read(self, address: int):
        con = self._con
        sim = self._sim

        addr = con.rdaddr
        data = con.rddata

        addr.araddr <<= address
        data.ready <<= False

        addr.valid <<= True
        await sim.true_on_clk(con.clk, addr.ready, timeout=10)
        addr.valid <<= False

        data.ready <<= True
        await sim.true_on_clk(con.clk, data.valid, timeout=10)
        data.ready <<= False

        return data.rdata.copy()

    async def read_multiple(self, addresses: list[int]):
        con = self._con
        sim = self._sim

        result = []

        addr = con.rdaddr
        data = con.rddata

        async def send_req():
            addr.valid <<= True
            for addr_val in addresses:
                addr.araddr.unsigned <<= addr_val
                await sim.true_on_clk(con.clk, addr.ready)
            addr.valid <<= False

        sim.start_soon(send_req())

        data.ready <<= True

        for _ in addresses:
            await sim.true_on_clk(con.clk, data.valid)
            result.append(data.rdata.copy())

        data.ready <<= False

        return result

    async def write(self, address: int, value: BitVector, wstrb: BitVector = Full):
        con = self._con
        sim = self._sim

        addr = con.wraddr
        data = con.wrdata
        resp = con.wrresp

        addr.awaddr <<= address
        addr.valid <<= True
        data.wdata <<= Unsigned[32](value)
        data.wstrb <<= wstrb
        data.valid <<= True
        resp.ready <<= False

        ack_addr = False
        ack_data = False

        while not (ack_addr and ack_data):
            await sim.clock_edge(con.clk)

            if not ack_addr and addr.ready:
                ack_addr = True
                addr.valid <<= False

            if not ack_data and data.ready:
                ack_data = True
                data.valid <<= False
                data.wdata <<= Null
                data.wstrb <<= Null

        resp.ready <<= True

        await sim.true_on_clk(con.clk, resp.valid, timeout=10)
        resp.ready <<= False

        await sim.clock_edge(con.clk)

        return resp.bresp.copy()

    async def write_multiple(self, addr_value: list[tuple[int, int]]):
        con = self._con
        sim = self._sim

        addr = con.wraddr
        data = con.wrdata
        resp = con.wrresp

        async def send_addr():
            for wraddr, _ in addr_value:
                addr.awaddr.unsigned <<= wraddr
                addr.valid <<= True
                await sim.true_on_clk(con.clk, addr.ready)

            addr.valid <<= False

        async def send_data():

            await sim.clock_cycles(con.clk, num_cycles=5)

            data.wstrb <<= Full
            for _, wrdata in addr_value:
                data.wdata.unsigned <<= wrdata
                data.valid <<= True
                await sim.true_on_clk(con.clk, data.ready)

            data.valid <<= False
            data.wstrb <<= Null

        result = []

        sim.start_soon(send_addr())
        sim.start_soon(send_data())

        resp.ready <<= True
        for _ in addr_value:
            await sim.true_on_clk(con.clk, resp.valid)
            result.append(resp.bresp.copy())

        resp.ready <<= False

        return result



from __future__ import annotations

from cohdl import Entity, Port, BitVector, Unsigned, Signed, Null
from cohdl import std

from pathlib import Path
from functools import partial
from ._build_simulation import prepare_ghdl_simulation
from ._proxy_port import ProxyPort
from .._base_simulation import _GenericParams, _BaseSimulator

import os

from cohdl_sim_ghdl_interface import GhdlInterface


class _Suspend:
    def __await__(self):
        yield None


_suspend = _Suspend()

global_alive_list = []


# ugly workaround to keep python objects referenced in the simulator alive
def _keep_alive(i):
    global_alive_list.append(i)
    return i


class Task:
    def __init__(self, simulator: Simulator):
        self._sim = simulator
        self._done = False
        self._continuation = None

    async def join(self):
        self._continuation = self._sim._current_coro
        while not self._done:
            await _suspend


class Simulator(_BaseSimulator):

    def _init_impl(self, p: _GenericParams):
        from cohdl_sim._build_cache import write_cache_file, load_cache_file

        assert (
            p.simulator == "ghdl"
        ), "cohdl_sim.ghdl_sim only supports the ghdl simulator"

        assert (
            p.sim_dir == p.vhdl_dir
        ), "cohdl_sim.ghdl_sim requires `sim_dir` and `vhdl_dir` to be the same"

        # ghdl_sim executes in the current context
        # set extra-env locally
        for name, val in p.extra_env.items():
            os.environ[name] = val

        top_name = p.entity._cohdl_info.name
        self._entity = p.entity
        self._top_name = top_name

        if not p.use_build_cache:
            lib = std.VhdlCompiler.to_vhdl_library(p.entity)
            vhdl_sources = (
                p.extra_vhdl_files + lib.write_dir(p.vhdl_dir) + p.extra_vhdl_files_post
            )
            write_cache_file(p.cache_file, p.entity, vhdl_sources=vhdl_sources)
        else:
            cache_content = load_cache_file(p.cache_file, p.entity)
            vhdl_sources = cache_content.vhdl_sources

        self._simlib = prepare_ghdl_simulation(
            vhdl_sources, top_name, p.sim_dir, copy_files=False
        )

        self._sim = GhdlInterface()
        self._sim_args = p.sim_args

        self._tb = None
        self._current_coro = None
        self._port_bv = p.cast_vectors

    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        simulator: str = "ghdl",
        sim_args: list[str] = None,
        sim_dir: str = "sim",
        vhdl_dir: str = "sim",
        cast_vectors=None,
        extra_env: dict[str, str] | None = None,
        extra_vhdl_files: list[str] = None,
        extra_vhdl_files_post: list[str] = None,
        use_build_cache: bool = False,
    ):
        p = _GenericParams(
            entity=entity,
            build_dir=build_dir,
            simulator=simulator,
            sim_args=sim_args,
            sim_dir=sim_dir,
            vhdl_dir=vhdl_dir,
            cast_vectors=cast_vectors,
            extra_env=extra_env,
            extra_vhdl_files=extra_vhdl_files,
            extra_vhdl_files_post=extra_vhdl_files_post,
            use_build_cache=use_build_cache,
        )

        super().__init__(p)
        self._init_impl(p)

    def _initial_fn(self):
        self._input_ports = {}
        self._output_ports = {}
        self._inout_ports = {}

        entity_name = self._entity.__name__

        class EntityProxy(self._entity):
            def __new__(cls):
                return object.__new__(cls)

            def __init__(self):
                pass

            def __str__(self):
                return entity_name

            def __repr__(self):
                return entity_name

        for name, port in self._entity._cohdl_info.ports.items():
            if self._port_bv is not None:
                port_type = type(Port.decay(port))

                if issubclass(port_type, BitVector) and not (
                    issubclass(port_type, (Signed, Unsigned))
                ):
                    if self._port_bv is Unsigned:
                        port = port.unsigned
                    elif self._port_bv is Signed:
                        port = port.signed
                    else:
                        raise AssertionError(
                            f"invalid default vector port type {self._port_bv}"
                        )

            proxy = ProxyPort(
                port,
                None,
                ghdl_handle=self._sim.handle_by_name(f"{self._top_name}.{name}"),
                sim=self,
            )

            setattr(EntityProxy, name, proxy)

            match port.direction():
                case Port.Direction.INPUT:
                    self._input_ports[name] = proxy
                case Port.Direction.OUTPUT:
                    self._output_ports[name] = proxy
                case Port.Direction.INOUT:
                    self._inout_ports[name] = proxy

        self._continue(self._tb(EntityProxy()))

    def _startup_function(self):
        # self._initial_fn()
        # global_alive_list.append(self._sim.add_startup_function(self._initial_fn))
        global_alive_list.append(self._sim.add_callback_delay(self._initial_fn, 0))

    def _continue(self, coro, name=None):
        prev_coro = self._current_coro
        self._current_coro = coro

        try:
            coro.send(None)
        except StopIteration:
            pass

        self._current_coro = prev_coro

    def test(self, testbench, sim_args=None):
        async def tb_wrapper(entity):
            await testbench(entity)
            self._sim.finish_simulation()

        sim_args = sim_args if sim_args is not None else self._sim_args

        self._tb = tb_wrapper
        self._sim.cleanup()

        global_alive_list.append(self._sim.add_startup_function(self._startup_function))
        self._sim.start(str(self._simlib), sim_args)
        self._sim.stop()

    async def _wait_picoseconds(self, picos: int):
        with _keep_alive(
            self._sim.add_callback_delay(
                partial(self._continue, self._current_coro, name="picos"), picos
            )
        ):
            await _suspend

    async def wait(self, duration: std.Duration):
        await self._wait_picoseconds(int(duration.picoseconds()))

    async def delta_step(self):
        with _keep_alive(
            self._sim.add_callback_delay(partial(self._continue, self._current_coro), 1)
        ):
            await _suspend

    async def rising_edge(self, signal: ProxyPort, /):
        with self._sim.add_callback_value_change(
            signal._root._handle,
            partial(self._continue, self._current_coro, name="rising_edge"),
        ):
            prev_state = signal.copy()

            while True:
                await _suspend
                new_state = signal.copy()

                if (not prev_state) and new_state:
                    return
                prev_state = new_state

    async def falling_edge(self, signal: ProxyPort, /):
        with self._sim.add_callback_value_change(
            signal._root._handle,
            partial(self._continue, self._current_coro, name="falling_edge"),
        ):
            prev_state = signal.copy()

            while True:
                await _suspend
                new_state = signal.copy()

                if prev_state and not new_state:
                    return
                prev_state = new_state

    async def any_edge(self, signal: ProxyPort, /):
        with self._sim.add_callback_value_change(
            signal._root._handle,
            partial(self._continue, self._current_coro, name="any_edge"),
        ):
            prev_state = signal.copy()

            while True:
                await _suspend
                new_state = signal.copy()

                if prev_state != new_state:
                    return
                prev_state = new_state

    async def clock_cycles(self, signal: ProxyPort, num_cycles: int, rising=True):
        if isinstance(signal, std.Clock):
            signal = signal.signal()

        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            prev_state = signal.copy()

            if rising:
                while True:
                    await _suspend
                    new_state = signal.copy()

                    if (not prev_state) and new_state:
                        num_cycles -= 1
                        if num_cycles == 0:
                            return

                    prev_state = new_state
            else:
                while True:
                    await _suspend
                    new_state = signal.copy()

                    if prev_state and not new_state:
                        num_cycles -= 1
                        if num_cycles == 0:
                            return

                    prev_state = new_state

    async def value_change(self, signal: ProxyPort):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            await _suspend
            return

    async def value_true(self, signal: ProxyPort):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            while not signal:
                await _suspend

    async def value_false(self, signal: ProxyPort):
        with self._sim.add_callback_value_change(
            signal._root._handle, partial(self._continue, self._current_coro)
        ):
            while signal:
                await _suspend

    async def start(self, coro):
        task = self.start_soon(coro)
        await self.delta_step()
        return task

    def start_soon(self, coro):

        task = Task(self)

        async def inner():
            await coro
            task._done = True

            if task._continuation is not None:
                self._continue(task._continuation)

        async def wrapper():
            await self.delta_step()
            self._continue(inner())

        self._continue(wrapper())

        return task

    def gen_clock(
        self,
        clk: ProxyPort,
        period_or_frequency: std.Duration = None,
        /,
        start_state=False,
    ):
        if isinstance(clk, std.Clock):
            if period_or_frequency is None:
                period_or_frequency = clk.frequency()

            clk = clk.signal()

        assert isinstance(period_or_frequency, (std.Frequency, std.Duration))

        period = period_or_frequency.period()

        half = int(period.picoseconds()) // 2

        async def thread():
            nonlocal clk

            while True:
                clk <<= start_state
                await self._wait_picoseconds(half)
                clk <<= not start_state
                await self._wait_picoseconds(half)

        self.start_soon(thread())



from setuptools import setup, Extension, find_packages
import os

ext_modules = []

if os.name == "posix":
    import pybind11
    import subprocess

    status, ghdl_result = subprocess.getstatusoutput(
        "ghdl --libghdl-include-dir",
    )

    if status != 0:
        print("determining ghdl include directory failed")
        print("continuing without direct ghdl bindings")
    else:
        ext_modules = [
            Extension(
                "cohdl_sim_ghdl_interface",
                [
                    "cpp/cohdl_sim_ghdl_interface/main.cpp",
                    "cpp/cohdl_sim_ghdl_interface/ghdl_cohdl_interface.cpp",
                ],
                include_dirs=[pybind11.get_include(), ghdl_result],
                extra_compile_args=["-O0", "-g3"],
            ),
        ]

setup(
    name="cohdl_sim",
    version="0.2.0",
    author="Alexander Forster",
    author_email="alexander.forster123@gmail.com",
    description="Simulation support library for CoHDL, based on cocotb",
    ext_modules=ext_modules,
    packages=find_packages(exclude=["examples"]),
)



from cohdl import Bit, BitVector, Signal

from cocotb.handle import Freeze, Release
from cocotb.triggers import RisingEdge, FallingEdge, Edge, ClockCycles
from ._base_proxy_port import _BaseProxyPort


class ProxyPort(_BaseProxyPort):
    def __init__(
        self,
        entity_port: Signal[BitVector],
        root=None,
        cocotb_port=None,
    ):
        super().__init__(entity_port, root)

        # only set in root port
        self._cocotb_port = cocotb_port

    def _load(self):
        val = self._cocotb_port.value

        if issubclass(self._type, (Bit, BitVector)):
            self._Wrapped._assign(val.binstr.upper())
        else:
            raise AssertionError(f"type {type(self._type)} not supported")

    def _store(self):
        if isinstance(self._Wrapped, Bit):
            self._cocotb_port.value = bool(self._Wrapped)
        else:
            self._cocotb_port.value = self._Wrapped.unsigned.to_int()

    def freeze(self):
        assert (
            self._is_root()
        ), "freeze may only be used on port objects, not on sub-bits/slices"
        self._cocotb_port.value = Freeze()

    def release(self):
        assert (
            self._is_root()
        ), "release may only be used on port objects, not on sub-bits/slices"
        self._cocotb_port.value = Release()

    async def _rising_edge(self):
        if self._is_root():
            await RisingEdge(self._cocotb_port)
        else:
            self._load()
            prev = bool(self._Wrapped)

            while True:
                await Edge(self._cocotb_port)

                self._load()
                current = bool(self._Wrapped)

                if current and not prev:
                    return
                prev = current

    async def _falling_edge(self):
        if self._is_root():
            await FallingEdge(self._cocotb_port)
        else:
            self._load()
            prev = bool(self._Wrapped)

            while True:
                await Edge(self._root._cocotb_port)

                self._load()
                current = bool(self._Wrapped)

                if prev and not current:
                    return
                prev = current

    async def _edge(self):
        if self._is_root():
            await Edge(self._cocotb_port)
        else:
            self._load()
            prev = self._Wrapped.copy()

            while True:
                await Edge(self._root._cocotb_port)

                self._load()
                current = self._Wrapped.copy()

                if prev != current:
                    return
                prev = current

    async def _clock_cycles(self, num_cycles, rising=True):
        if self._is_root():
            await ClockCycles(self._cocotb_port, num_cycles, rising)
        else:
            if rising:
                for _ in num_cycles:
                    await self._rising_edge()
            else:
                for _ in num_cycles:
                    await self._falling_edge()

    def __await__(self):
        async def gen():
            while True:
                if self:
                    return
                await self._edge()

        return gen().__await__()



import cohdl
from cohdl import std, Bit, BitVector, Unsigned, Signed, Port, Entity

from dataclasses import dataclass
from pathlib import Path
import json


def _store_port_info(port: cohdl.Port):
    port_type = std.base_type(port)

    dir = port.direction().name

    if issubclass(port_type, Bit):
        return {"type": "Bit", "dir": dir}
    if issubclass(port_type, cohdl.Boolean):
        return {"type": "Bool", "dir": dir}
    if issubclass(port_type, BitVector):
        w = port_type.width

        if issubclass(port_type, Unsigned):
            return {"type": "Unsigned", "width": w, "dir": dir}
        if issubclass(port_type, Unsigned):
            return {"type": "Signed", "width": w, "dir": dir}

        return {"type": "BitVector", "width": w, "dir": dir}

    raise AssertionError(f"invalid '{port_type=}'")


def _load_port_info(info: dict):
    dir = Port.Direction[info["dir"]]

    match info["type"]:
        case "Bit":
            t = Bit
        case "Bool":
            t = cohdl.Boolean
        case "Unsigned":
            t = Unsigned[info["width"]]
        case "Signed":
            t = Signed[info["width"]]
        case "BitVector":
            t = BitVector[info["width"]]
        case _:
            raise AssertionError(f"invalid '{info['type']=}'")

    return Port[t, dir]()


@dataclass
class CacheContent:
    vhdl_sources: list[str]


def write_cache_file(path: Path, entity: type[Entity], vhdl_sources: list[str]):
    top_ports = {
        name: _store_port_info(port) for name, port in entity._cohdl_info.ports.items()
    }

    with open(path, "w") as cache:
        json.dump(
            {"vhdl_sources": vhdl_sources, "top_ports": top_ports}, cache, indent=2
        )


def load_cache_file(path: Path, entity: type[Entity]) -> CacheContent:
    with open(path) as cache_file:
        cache = json.load(cache_file)

    vhdl_sources = cache["vhdl_sources"]
    top_ports = cache["top_ports"]

    for name, info in top_ports.items():
        if not hasattr(entity, name):
            std.add_entity_port(entity, _load_port_info(info), name=name)

    return CacheContent(vhdl_sources=vhdl_sources)



from cohdl import Entity

from .._base_simulation import _BaseSimulator

class Simulator(_BaseSimulator):
    def __init__(
        self,
        entity: type[Entity],
        *,
        build_dir: str = "build",
        sim_dir: str = "sim",
        vhdl_dir: str = "sim",
        extra_vhdl_files: list[str] = None,
        extra_vhdl_files_post: list[str] = None,
        cast_vectors=None,
        simulator: str = "ghdl",
        sim_args: list[str] | None = None,
        extra_env: dict[str, str] | None = None,
    ):
        """
        This is an alternative simulator that directly invokes GHDL without cocotb.

        The main advantage over the default simulator is that GHDL is loaded as a
        library into the current process. That makes it possible to use the Python
        debugger to set breakpoint and step threw testbench code. You can also define tests in
        Jupyter notebooks.

        ghdl_sim.Simulator is only available on Linux systems. ghdl must be installed
        before the cohdl_sim package because the FFI library is build during installation.
        In addition the ghdl-gcc backend in required.

        Differences to the default simulator:

        * `sim_dir` must be equal to `vhdl_dir`
        * `extra_env` is set in the local process environment
        """
